"""Rebuild -- rebuild droplet timeseries from run records.
This file is part of the dropsignal package.
Copyright 2016-2019 Guilhem Doulcier, Licence GNU AGPL3+
"""
import logging

from itertools import combinations
from collections import defaultdict

import numpy as np
import pandas as pd

logger = logging.getLogger('dropsignal')

def rebuild(runs, peaks, n_drops, coalescence_events, relative_times:bool):
    """Rebuild the droplet timeseries from runs records."""

    # Read the coalescence informations
    jumps, id_mappings = coalescence_indices(runs.id, n_drops, coalescence_events)

    # Concatenate the dataframes
    droplet_dyn = droplet_dynamics(runs[runs.good].id, peaks, id_mappings)

    # Look in which runs we have complete coalescence information
    for r, n_peaks in runs.set_index('id').peaks.iteritems():
        print('run', r, n_peaks, len(jumps[r]))
    runs['complete_coalescence_information'] = [n_drops-len(jumps[r]) == n_peaks
                                                for r,n_peaks
                                                in runs.set_index('id').peaks.iteritems()]

    # The sorting run is the last one for wich we have complete coalescence information.
    last_complete_info = runs[runs.complete_coalescence_information].id.max()

    logger.info(("{}/{} runs with complete coalescence information. Last one is {}"
                 "").format(runs.complete_coalescence_information.sum(), runs.good.sum(), last_complete_info))

    # Set the time columns
    droplet_dyn['absolute_time'] = droplet_dyn['time']
    if relative_times:
        logger.debug('Using relative times as default time')
        if droplet_dyn['relative_time'].isnull().any():
            logger.error('Failed to compute relative time from droplet creation. Fix the detection parameters or set "relative_times" to `False` in the protocol.')
            raise ValueError('Failed to compute relative time from droplet creation.')
        droplet_dyn['time'] = droplet_dyn['relative_time']
    else:
        logger.debug('Using absolute times as default time')
    return droplet_dyn, runs, jumps, id_mappings, last_complete_info

def add_jumps(jumps, new_jumps):
    """Add new jumps to the list of jumps.

    Args:
        jumps (list): existing jumps
        new_jumps (list): jumps to add

    For each new jump we have to reduce by one the index of the existing
    jumps that occurs later in order to maintain consistency.
    """
    for new_jump in new_jumps:
        for j, jump in enumerate(jumps):
            if new_jump < jump:
                jumps[j] -= 1
        jumps.append(new_jump)
    return jumps

def mappings_id_exp_to_run(n_drops, jumps):
    """Build the mappings id_run <-> id_exp.

    Args:
        n_drops (int): number of droplets
        jumps (list): list of id_exp to jumps.
    """
    id_exp = np.arange(n_drops)
    id_run = np.arange(n_drops)
    for jump in jumps:
        id_exp[id_run > jump] += 1
    return {'exp_to_run':{e:r for e,r in zip(id_exp, id_run)},
            'run_to_exp':{r:e for e,r in zip(id_exp, id_run)}}

def coalescence_indices(runs, n_drops, manual_coalescence_events):
    """

    Coalescence events is a mapping {run: list_of_pairs}.
    Each pairs in the `list_of_pairs` are the peak_id (1-indexed)
    of the two droplets that coalesced during the run `run+1`
    """
    jumps = {runs.min()-1:[]}
    coalescence_events = defaultdict(list)

    # Read the user input.
    # Convert the peak_id to a id_run
    for (run, id_peak1, id_peak2) in manual_coalescence_events:
        coalescence_events[run].append(min(id_peak1-1,id_peak2-1))

    for run in runs:
        new_jumps = coalescence_events[run]
        jumps[run] = add_jumps(jumps[run-1].copy(), new_jumps)
        logger.debug('Run {} - peak jumps are {}'.format(run, jumps[run]))

    id_mappings = {run: mappings_id_exp_to_run(n_drops, jumps[run]) for run in runs}
    return jumps, id_mappings


def concatenate_channels(peaks_run, columns=('mean', 'max','shift', 'std','cov','median','area')):
    """Concatenate a dictionnary of channels
           Args:
               peaks_runs (dict): Contains a dataframe by channel.
               columns (iterable): Columns specific to the channel.
           Returns: A pd.Dataframe with a <channel>_<col> column per <channel> and <columns>
               and using the others columns as index.
    """
    dyn = []
    if len(peaks_run)>1:
        for channel, peak in peaks_run.items():
            if channel != '__measure__':
                dyn.append(peak.rename(columns=dict([(col, channel+"_"+col) for col in columns if col in peak.columns])))
                idx = [x for x in dyn[-1].columns if "_".join(x.split("_")[:-1]) != channel]
                dyn[-1] = dyn[-1].set_index(idx)
        df = pd.concat(dyn, 1)
        df.reset_index(inplace=True)
        # Add the columns "time", "speed", "size" from the __measure__ channel.
        if '__measure__' in peaks_run.keys():
            df = pd.merge(df, peaks_run['__measure__'], left_on='id_run', right_on='id_run')

        if "Time_value" in df.columns:
            df.rename(columns={"Time_value":"time"}, inplace=True)
        return df
    else:
        # Run 0 does not have the measures.
        return peaks_run['__measure__']

def droplet_dynamics(runs, peaks, id_mappings):
    """ Concatenate the runs in a single data frame

    Args:
        runs (iterable): Id of the runs to use
        peaks (iterable of dict): Each dictionnary correspond to a run,
            each one contains a dataframe by channel with columns: max, mean, value,
            center, id_run, time.
    Returns:
        A single dataframe with:
           - center, id_run, time and run from the runs,
           - CHANNEL_max, CHANNEL_mean, CHANNEL_value from each channel
           - id_exp and offset according to the coalescence events.
    """
    droplet_dyn = []
    for run in runs:
        # We start by concatenating all channels.
        dyn = concatenate_channels(peaks[run])
        dyn['run'] = run

        # Apply index the mapping.
        dyn['id_exp'] = [id_mappings[run]['run_to_exp'][id_run] for id_run in dyn.id_run]
        dyn['offset'] = 0
        droplet_dyn.append(dyn)

    droplet_dyn = pd.concat(droplet_dyn).reset_index(drop=True)

    logger.info('Computing time relative to droplet creation')
    droplet_dyn['relative_time'] = np.nan
    try:
        for _, data in droplet_dyn.groupby('id_exp'):
            relative_times = data['time'] -  data[data.run==0].time.values[0]
            droplet_dyn.loc[relative_times.index,'relative_time'] = relative_times.values
    except Exception:
        droplet_dyn['relative_time'] = np.nan
        logger.critical('Failed to compute times relative to droplet creation')
    return droplet_dyn

def find_coalescent_position(previous, current):
    """Locate coalescence events between two run by looking at droplet size.

    Args:
        previous (dict): data from previous run.
        current (dict): data from current run.

    Returns a list of indices. There will be one indice per
    coalescence event, i.e number of peak in current run minus the
    number of peak in the previous one.

    This function try all possible combination of coalescence event
    position (N choose C, where N is the number of drops, C the number
    of events) and for each compute the difference in droplet size
    between the two runs (ignoring the coalesced ones). The position
    returned is the one for which this quantity is minimal.

    Inputs `previous` and `current` must contain a dataframe under the
    key "__measure__", whith one line per peak. Droplet size will be
    taken from the column `size`, or `measure_points` if `size` is
    missing. The only other required column is `id_run`.
    """

    column = 'size'

    if '__measure__' not in previous.keys() or '__measure__' not in current.keys():
        logger.warning('Photodiode data missing. Impossible to find the position of the coalesced droplet.')
        return None

    if 'size' not in previous['__measure__'].columns or 'size' not in current['__measure__'].columns:
        logger.warning('Droplet size data missing. Trying to use the number of measured points.')
        column = 'measure_points'

        if column not in previous['__measure__'].columns or column not in current['__measure__'].columns:
            logger.warning('Number of measured points missing. Impossible to find the position of the coalesced droplet.')
            return None

    # Align the two runs as much as possible by matching id_runs.
    merged = pd.merge(previous['__measure__'].loc[:, ['id_run', column]],
                      current["__measure__"].loc[:, ['id_run', column]],
                      how='left', left_on='id_run', right_on='id_run',
                      suffixes=['_prev', '_curr'])

    # Count the number of coalescence events
    current_size = merged[column+'_curr'].dropna().values
    number_coal = merged.shape[0] - len(current_size)

    # Compute all the difference in size for different offset.
    diff = {d:np.abs(current_size-merged[column+'_prev'].values[d:-number_coal+d if -number_coal+d<0 else None])
            for d in np.arange(number_coal+1)}

    # Compute all the positions possible for the coalescence events
    keys = list(combinations(np.arange(merged.shape[0]), number_coal))

    # Compute all the total difference in size for all different coalescence events positions.
    delta = np.zeros(len(keys))
    for i,k in enumerate(keys):
        for delay, (start, end) in enumerate(zip([0]+list(k),list(k)+[None])):
            delta[i] += diff[delay][start:end].sum()

    # Get back the position of the drop.
    pos = [merged.ix[i,'id_run'] for i in keys[np.argmin(delta)]]
    reason = 'Minimal droplet size total variation: {} | position of breakpoints: {}'.format(np.min(delta), keys[np.argmin(delta)])
    return pos, reason


def get_coalescence_suggestions(runs, peaks):
    coalescence_events = []
    coalescent_runs = [runs.id[k]
                       for k in runs.index
                       if 'coalescence_event' in frozenset(runs.tags[k])]
    for run in coalescent_runs:
        try:
            candidate,reason = find_coalescent_position(peaks[run-1],peaks[run])
        except Exception as ex:
            logger.exception("Error in coalescence detection  between run {} and {}".format(run-1, run))
            candidate = None
        if candidate != None:
            for c in candidate:
                coalescence_events.append((run, c, 'auto'))
            logger.info('Between run {} and {} best candidate is drop {} ({}).'.format(run-1,
                                                                                       run,
                                                                                       candidate,
                                                                                       reason))
    logger.info('{} coalescence event(s) detected in {} suspicious runs.'.format(len(coalescence_events),
                                                                                 len(coalescent_runs)))
    return coalescence_events
